////////////////////////////////////////////////////////
// wddxPipe.js - Helper functions for urlPipe applet
// Author: Nate Weiss, April 1999, for WDDX SDK Beta 2
// Parameters for the various functions defined herein:
// - pipeApplet: Instance name of a urlPipe applet
// - url:        The url to fetch from
// - params:     Object of name/value pairs to post to url
////////////////////////////////////////////////////////



////////////////////////////////////////////////////////
// This function fetches a raw web page via a urlPipe instance
function fetchPageFromWeb(pipeApplet, url, params) {

  // If URL does not start with "http", 
  // Assume a path based on current folder
  if (url.indexOf('http') != 0) {
    var ThisURL = window.location.href;
    var BaseURL = ThisURL.substring(0, ThisURL.lastIndexOf('/'));
    url = BaseURL + '/' + url;
  }  
  
  // If no params object was provided, create one with a dummy param
  // This is to work around "hang if no parameters" problem with urlPipe
  if (params == null) {
    params = new Object;
    params["DummyParameter"] = 1;
  } 
  

  // For each parameter, call urlPipe's addParam function
  for (prop in params) {
    pipeApplet.addParam(prop, escape(params[prop]));
  }
  
  // Fetch the actual packet
  pipeApplet.catchPage = true;
  pipeApplet.post(url);


  // Return page content to calling code
  return new String(pipeApplet.getPage());
}



////////////////////////////////////////////////////////
// This function fetches a WDDX Packet from a web page
// It calls fetchPageFromWeb internally
function fetchPacketFromWeb(pipeApplet, url, params) {

  // Delimiters to consider the start and end of packet
  var tokStart = '<wddxPacket';
  var tokEnd = '</wddxPacket>';
  var fetched, posStart, posEnd;
  var fetchedPacket = '';
  var thisRetry = 0;

  // Number of retries to attempt
  var Retries = 5;
  if (arguments.length > 3) Retries = arguments[3];

  // Attempt packet-fetching up to the number of retries
  while ( (thisRetry <= Retries) && (fetchedPacket == '') ) {
      
    // Fetch raw web page, using the urlPipe applet
    fetched = fetchPageFromWeb(pipeApplet, url, params);
  
    // Where does packet begin and end within page?  
    posStart = fetched.indexOf(tokStart);
    posEnd = fetched.indexOf(tokEnd, posStart);
  
    // We'll return packet if found; otherwise return null
    if (posEnd > posStart) 
      fetchedPacket = fetched.substring( posStart, (posEnd + tokEnd.length) );

    // increment the retry counter      
    thisRetry++
  }	

  // Return finished packet to calling code    
  return fetchedPacket;
}



////////////////////////////////////////////////////////
// This function fetches a recordset from a web page
// It calls fetchPacketFromWeb internally
function fetchRecordsetFromWeb(pipeApplet, url, params) {

  // Number of retries to attempt
  var Retries = 5;
  if (arguments.length > 3) Retries = arguments[3];

  // Fetch WDDX Packet from web page
  var Packet = fetchPacketFromWeb(pipeApplet, url, params, Retries);
  if (Packet == '') return null;

  // Create new recordset and populate from packet
  var TempRS = new WddxRecordset;
  TempRS.readFromPacket(Packet);
  
  // Return new recordset to calling code
  return TempRS;  
}


